
import time
import math
import random
import sys

import pygame
from pygame.locals import *
from OpenGL.GL import *

from define import *
from StateMenu import StateMenu
from Texture import Texture
from Config import Config


# Core game class, handles general initialization and the main game loop. 
# Handles keyboard/mouse input and performs the step function of the current
# game State.
class Core:


    # Keyboard layout. Currently supports both QWERTY and Dvorak.
    LAYOUT_KEYBOARD = {
            K_ESCAPE: KEY_ESC,
            K_RETURN: KEY_ENTER,
            K_a: KEY_LEFT,
            K_COMMA: KEY_UP,
            K_w: KEY_UP,
            K_e: KEY_RIGHT,
            K_d: KEY_RIGHT,
            K_o: KEY_DOWN,
            K_s: KEY_DOWN,
            K_TAB: KEY_MAP,
            #K_BACKSPACE: KEY_CHEAT,
            K_1: KEY_1,
            K_2: KEY_2,
            K_3: KEY_3,
            K_4: KEY_4,
            K_5: KEY_5}


    # Mouse layout, since the mouse is only used for firing, this is pretty 
    # simple.
    LAYOUT_MOUSE = {
            1: KEY_FIRE}


    def __init__(self):

        # Read config
        self.config = Config(FN_CFG)

        # Init Pygame and OpenGL

        # Setting a small buffer size to reduce sound latency
        pygame.mixer.pre_init(22050, -16, 2, 512) 

        pygame.init()
        self.resizeDisplay([self.config.view_xx, self.config.view_yy])
        glEnable(GL_TEXTURE_2D)
        glEnable(GL_CULL_FACE)
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        glClear(GL_COLOR_BUFFER_BIT);

        glEnableClientState(GL_VERTEX_ARRAY)
        glEnableClientState(GL_TEXTURE_COORD_ARRAY)
        glEnableClientState(GL_COLOR_ARRAY)

        # Lists containing the key and mouse states
        # Note that these will be passed to the states, so the pointers should 
        # not be changed.
        self.keys = [0]*KEY_N
        self.mouse = [0, 0]

        # Several States will need the Hud texture, the other textures are 
        # only used by StateMain and initialized there.
        self.tex_hud = Texture('hud.png')

        # State queue, only the last state on the queue will be used.
        self.states = []
        self.states.append(StateMenu(self))

        self.loop()


    # Main game loop
    # The game uses steps as unit, which is the smallest game state update. 
    def loop(self):
        clock = pygame.time.Clock()
        while True:
            self.updateInput()
            self.states[-1].step()
            clock.tick(STEP_FREQ)


    # Reads the input into self.keys. 
    # Value X of self.keys[KEY] indicates:
    # 0: Key not held down
    # 1: Key currently held down
    # 2: Key released since last step
    # 3: Key pressed since last step
    # self.mouse stores the relative movement since the last step (this is a 
    # FPS after all)
    def updateInput(self):
        for i in range(KEY_N):
            if (self.keys[i] >= 2):
                self.keys[i] -= 2
        self.mouse[0] = 0
        self.mouse[1] = 0
        for event in pygame.event.get():
            if event.type == QUIT:
                sys.exit(0)
            elif event.type == KEYDOWN:
                if event.key in self.LAYOUT_KEYBOARD:
                    self.keys[self.LAYOUT_KEYBOARD[event.key]] = 3
            elif event.type == KEYUP:
                if event.key in self.LAYOUT_KEYBOARD:
                    self.keys[self.LAYOUT_KEYBOARD[event.key]] = 2
            elif event.type == MOUSEBUTTONDOWN:
                if event.button in self.LAYOUT_MOUSE:
                    self.keys[self.LAYOUT_MOUSE[event.button]] = 3
            elif event.type == MOUSEBUTTONUP:
                if event.button in self.LAYOUT_MOUSE:
                    self.keys[self.LAYOUT_MOUSE[event.button]] = 2
            elif event.type == MOUSEMOTION:
                self.mouse[0] += event.rel[0]
                self.mouse[1] += event.rel[1]
            elif event.type == VIDEORESIZE:
                self.resizeDisplay(event.size)


    # Recompute Frustum matrixy stuff.
    def resizeDisplay(self, scr_xx):
        self.scr_xx = scr_xx
        self.scr = pygame.display.set_mode(scr_xx, 
                pygame.OPENGL | pygame.DOUBLEBUF | pygame.RESIZABLE)
        glViewport(0, 0, scr_xx[0], scr_xx[1])
        znear = 1.0/16
        zfar = 1024.0
        xfov = 90.0
        aspect = float(scr_xx[0])/scr_xx[1]
        xmax = znear*math.tan(xfov*math.pi/360)
        ymax = xmax/aspect
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        glFrustum(-xmax, xmax, -ymax, ymax, znear, zfar)
        glMatrixMode(GL_MODELVIEW)


    def screenshot(self, i):
        fn = 'stuff/scr/' + str(i).zfill(8) + '.png'
        print fn
        pygame.image.save(self.scr, fn)

