
import math
import random
import copy
import os

from OpenGL.GL import *

from define import *
from Tile import Tile
from Texture import Texture
from glarray import *

# Imports for every thing that can be created on loading:
from ThingStatue import ThingStatue
from ThingKey import ThingKey
from ThingDoor import ThingDoor
from ThingEntrance import ThingEntrance
from ThingExit import ThingExit
from ThingWave import ThingWave
from ThingGhost import ThingGhost
from ThingSkeleton import ThingSkeleton
from ThingWeaponLaser import ThingWeaponLaser
from ThingWeaponBounce import ThingWeaponBounce
from ThingWeaponSpread import ThingWeaponSpread
from ThingWeaponTv import ThingWeaponTv
from ThingFlower import ThingFlower
from ThingTriggerBg import ThingTriggerBg
from ThingClosed import ThingClosed
from ThingTriggerFlag import ThingTriggerFlag
from ThingDragon import ThingDragon
from ThingHeart import ThingHeart
from ThingShadow import ThingShadow
from ThingStatueAppear import ThingStatueAppear
from ThingBoss import ThingBoss
from ThingConfigMusic import ThingConfigMusic
from ThingConfigMouse import ThingConfigMouse
from ThingConfigDiff import ThingConfigDiff
from ThingFace import ThingFace


# Game map, containing data for a single level.
class Map:


    # Dictionary storing the creation function and arguments for each Thing.
    # The arguments are read from the .lev files and passed as strings.
    THINGCREATE = {
            'entrance': (ThingEntrance, 3),
            'exit': (ThingExit, 2),
            'statue': (ThingStatue, 4),
            'key': (ThingKey, 3),
            'door': (ThingDoor, 3),
            'wave': (ThingWave, 2),
            'ghost': (ThingGhost, 3),
            'skeleton': (ThingSkeleton, 3),
            'weapon_laser': (ThingWeaponLaser, 2),
            'weapon_bounce': (ThingWeaponBounce, 2),
            'weapon_spread': (ThingWeaponSpread, 2),
            'weapon_tv': (ThingWeaponTv, 2),
            'flower': (ThingFlower, 2),
            'trigger_bg': (ThingTriggerBg, 3),
            'closed': (ThingClosed, 4),
            'trigger_flag': (ThingTriggerFlag, 3),
            'dragon': (ThingDragon, 3),
            'heart': (ThingHeart, 2),
            'shadow': (ThingShadow, 2),
            'statue_appear': (ThingStatueAppear, 5),
            'boss': (ThingBoss, 3),
            'cfg_music': (ThingConfigMusic, 2),
            'cfg_mouse': (ThingConfigMouse, 2),
            'cfg_diff': (ThingConfigDiff, 2),
            'face': (ThingFace, 3)}


    def __init__(self, main, filename):
        self.main = main
        self.size = (32, 32)
        self.tiles = [None]*self.size[0]*self.size[1] # Actual map data array
        self.start = (0,0,0) # Player start position, (x,y,dir)
        self.dl = glGenLists(1)
        self.filename = filename 
        self.tex_bg = None # Background texture
        self.flags = set()
        self.music = None


    def __del__(self):
        glDeleteLists(self.dl, 1)


    # The load function is separated from the creation function since some of 
    # the constructors called here might require self.main.map, which isn't 
    # set correctly until after the Map constructor.
    def load(self):
        f = open(os.path.join(DATA_DIR, self.filename))
        data = f.read()
        f.close()

        data = data.split()
        i = 0
        # The tile dictionary maps tile tags to their actual tiles. 
        tiledict = dict()
        while i < len(data):
            # Note that in order for things to work right, the 'tile' and 
            # 'size' tags should happen before 'map', and the thing tags 
            # should be placed after 'map'.
            if data[i] == 'tile':
                # Tile definition, given as:
                # tag: (solidity, (visual) tile type, texture_x, texture_y)
                tiledict[data[i+1]] = Tile(self.main,
                        int(data[i+2]), data[i+3],
                        (int(data[i+4]), int(data[i+5])))
                i += 6
            elif data[i] == 'size':
                # Map size.
                self.size = (int(data[i+1]), int(data[i+2]))
                i += 3
            elif data[i] == 'map':
                # Map data, reads the map tags and creates the corresponding
                # Tile objects.
                self.tiles = [copy.copy(tiledict[j]) for j in 
                        data[i+1:i+1+(self.size[0]*self.size[1])]]
                i += 1+self.size[0]*self.size[1]
            elif data[i] == 'start':
                # Player start position, does not call an actual thing 
                # constructor directly.
                self.start = (int(data[i+1]), int(data[i+2]), 
                        float(data[i+3])/180.0*math.pi)
                self.get(self.start).spr = (1,2)
                i += 4
            elif data[i] == 'background':
                # Background file
                self.tex_bg = Texture(data[i+1])
                i += 2
            elif data[i] == 'music':
                # Music file
                self.music = data[i+1]
                i += 2
            elif data[i] in self.THINGCREATE:
                # Thing definition, calls thing constructor with its arguments.
                initfun = self.THINGCREATE[data[i]][0]
                argc = self.THINGCREATE[data[i]][1]
                args = [data[i+j+1] for j in range(argc)]
                initfun(self.main, args)
                i += argc+1
            else:
                print('error: unknown map tag: ' + data[i])
                i += 1
        self.recompileDl()


    # Compile the displaylist for the map landscape. Changes to the map will 
    # not be visible unless this is called again (and this takes a while, so 
    # it's probably a better idea to handle such things differently).
    def recompileDl(self):
        vertices = []
        texcoords = []
        colors = []
        glEnableClientState(GL_COLOR_ARRAY)
        for j in range(self.size[1]):
            for i in range(self.size[0]):
                self.get((i,j)).draw(self,(i,j),vertices,texcoords,colors)
        count = len(vertices)/3
        vertices = glarray(vertices)
        texcoords = glarray(texcoords)
        colors = glarray(colors)
        glVertexPointer(3,GL_FLOAT,0,vertices)
        glTexCoordPointer(2,GL_FLOAT,0,texcoords)
        glColorPointer(3,GL_FLOAT,0,colors)

        glNewList(self.dl, GL_COMPILE)
        glDrawArrays(GL_QUADS,0,count)
        glEndList()


    def draw(self):
        glCallList(self.dl)


    def drawBg(self):
        # Note that if there is no background, this should be skipped in 
        # StateMain
        ytop = float(self.main.core.scr_xx[1])/self.main.core.scr_xx[0]
        vertices = glarray([-1, 0, -1, 1, 0, -1, 1, ytop, -1, -1, ytop, -1])
        off = -self.main.cam[3]/360.0
        texcoords = glarray([off, 1, off+0.25, 1, off+0.25, 0, off, 0])
        glVertexPointer(3,GL_FLOAT,0,vertices)
        glTexCoordPointer(2,GL_FLOAT,0,texcoords)
        glDisableClientState(GL_COLOR_ARRAY)
        glDisable(GL_DEPTH_TEST)
        glPushMatrix()
        glLoadIdentity()
        glDrawArrays(GL_QUADS,0,4)
        glPopMatrix()
        glEnable(GL_DEPTH_TEST)
        glEnableClientState(GL_COLOR_ARRAY)


    def get(self, pos):
        return self.tiles[pos[0] + pos[1]*self.size[0]]


    def safeGet(self, pos):
        pos = list(pos)
        pos[0] = bound(pos[0], 0, self.size[0]-1)
        pos[1] = bound(pos[1], 0, self.size[1]-1)
        return self.tiles[pos[0] + pos[1]*self.size[0]]


    def getSolid(self, pos): 
        return self.safeGet(pos).solid


    def getHeight(self, pos):
        return self.safeGet(pos).height






