
import random
import math
import os

import pygame
from pygame.locals import *
from OpenGL.GL import *

from glarray import *
from define import *
from Map import Map
from ThingPc import ThingPc
from StatePause import StatePause
from Texture import Texture
from Hud import Hud
from Audio import Audio
from StateEndgame import StateEndgame


# Main game state, handles the main FPS gameplay.
class StateMain:



    def __init__(self, core, save):
        self.core = core
        self.keys = self.core.keys
        self.mouse = self.core.mouse
        pygame.mouse.set_visible(False)
        pygame.event.set_grab(True)
        self.hud = Hud(self)

        # List of Things, or game entities.
        self.things = []
        # All solid things are stored in this set as well, for faster 
        # collision checking.
        self.things_solid = set()
        # Set of all destructible things:
        self.things_destr = set()
        # Set of registered things, used for various purposes.
        self.things_reg = set()
        # Pc thing
        self.pc = None
        # Boss thing, if it exists
        self.boss = None
 
        # Load textures
        self.tex_things = Texture('things.png')
        self.tex_landscape = Texture('landscape.png')
        self.tex_hud = self.core.tex_hud

        # Init audio system
        self.audio = Audio(self)

        # List of sprites to be drawn each step.
        self.sprites = []

        # List of sounds to be played each step.
        self.sounds = []

        # Camera position (x,y,z,p,q), where p and q indicate the direction.
        self.cam = (0.0, 0.0, 0.0, 0.0, 0.0)

        # Step counter.
        self.step_i = 0

        # Color of flash to overlay on the screen, uses transparency, so if the 
        # fourth value is 0.0 it won't be visible.
        self.flash = [0.0]*4

        # Name of next map, if this is set then the map will change at the end
        # of the step.
        self.nextmap = ''

        # Keys the PC has found, multiple keys of type 0 can be collected, 
        # these are hubkeys. For type 1-4 we just fill in 1 or 0 depending on 
        # whether the key has been collected.
        # An unused hubkey is a tuple (-1,-1), a used hubkey is a tuple 
        # representing the coordinates at which its door is.
        self.mapkeys = [save['hubkeys']] + [0]*4

        # Levels which are already finished (in which the PC has left through 
        # an ThingExit). Finished maps can not be revisited.
        self.finmaps = save['finmaps']

        # Whether hard mode is toggled
        self.hardmode = save['hard']

        # Initialize ammo
        self.ammo = save['ammo']
        self.ammo[0] = 1
        self.prevammo = list(self.ammo) # Ammo at time of level start
        self.prevhubkeys = list(self.mapkeys[0])
        self.weapon_i = WEAPON_RAKE # Weapon selection

        # Last PC position and direction in hub.
        self.hubpc = {'pos': None, 'dir': None}

        # Whether cheat mode is toggled.
        self.cheat = False

        # Enable drawing fog, for better vision depth.
        glEnable(GL_DEPTH_TEST)
        glFogi(GL_FOG_MODE,GL_EXP2)
        glFogf(GL_FOG_DENSITY,1.0/32)
        glFogfv(GL_FOG_COLOR,[0.0]*4);
        glEnable(GL_FOG)
        glEnableClientState(GL_COLOR_ARRAY)

        # Create flash display list
        self.flashdl = glGenLists(1)
        vertices = glarray([-16, -16, -1, 16, -16, -1, 16, 16, -1, -16, 16, -1])
        glVertexPointer(3,GL_FLOAT,0,vertices)
        glDisableClientState(GL_COLOR_ARRAY)
        glDisableClientState(GL_TEXTURE_COORD_ARRAY)
        glNewList(self.flashdl, GL_COMPILE)
        glDisable(GL_TEXTURE_2D)
        glDisable(GL_DEPTH_TEST)
        glPushMatrix()
        glLoadIdentity()
        glDrawArrays(GL_QUADS,0,4)
        glPopMatrix()
        glEnable(GL_TEXTURE_2D)
        glEnable(GL_DEPTH_TEST)
        glEndList()
        glEnableClientState(GL_COLOR_ARRAY)
        glEnableClientState(GL_TEXTURE_COORD_ARRAY)

        # Load the first map. 
        self.loadMap(LEVEL_HUB)


    def __del__(self):
        glDeleteLists(self.flashdl, 1)


    def step(self):
        self.processStep()
        self.drawScreen()
        self.playSounds()
        if SCREENCAP > 0 and self.step_i % SCREENCAP == 0:
            self.core.screenshot(self.step_i)
        self.step_i += 1
        if self.nextmap != '':
            # Switch to next map if necessary
            if self.nextmap == 'endgame':
                self.audio.playMusic(None)
                self.core.states.append(StateEndgame(self.core))
            else:
                self.loadMap(self.nextmap)
                self.nextmap = ''
        if self.pc.nrg <= 0:
            self.loadMap(self.map.filename)


    def processStep(self):

        # Pausing the game
        if self.keys[KEY_ESC] == 3:
            pygame.mixer.pause()
            pygame.mixer.music.pause()
            self.core.states.append(StatePause(self.core))

        # Cheat mode
        if self.keys[KEY_CHEAT] == 3:
            self.cheat ^= True
            print 'cheat mode:', self.cheat
            if self.cheat:
                glDisable(GL_FOG)
            else:
                glEnable(GL_FOG)


        # Perform step function for all Things.
        for t in self.things:
            t.step()

        # Delete all things with no energy left.
        # TODO set filter function?
        self.things = [t for t in self.things if t.nrg > 0]
        self.things_solid = set([t for t in self.things_solid if t.nrg > 0])
        self.things_destr = set([t for t in self.things_destr if t.nrg > 0])
        self.things_reg = set([t for t in self.things_reg if t.nrg > 0])

        # Print some debugging output.
        #if self.step_i % STEP_FREQ == 0: 
        #    print [int(i) for i in self.pc.pos[:2]]


    def drawScreen(self):
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

        glPushMatrix()

        # Draw background
        if self.map.tex_bg != None:
            glColor4f(1.0, 1.0, 1.0, 1.0)
            self.map.tex_bg.bind()
            self.map.drawBg()

        # Translate/rotate according to the camera position and direction.
        glRotatef(-90.0, 1.0, 0.0, 0.0)
        glRotatef(-self.cam[4], 1.0, 0.0, 0.0)
        glRotatef(-self.cam[3]+90.0, 0.0, 0.0, 1.0)
        glTranslatef(-self.cam[0], -self.cam[1], -self.cam[2])

        # Draw the map
        self.tex_landscape.bind()
        glEnableClientState(GL_COLOR_ARRAY)
        self.map.draw()
        glDisableClientState(GL_COLOR_ARRAY)

        # Draw the sprites
        glColor3f(1.0, 1.0, 1.0)
        self.tex_things.bind()
        self.drawThings()

        # Draw the hud
        self.tex_hud.bind()
        self.hud.draw()

        # Draw a flash effect
        glDisableClientState(GL_TEXTURE_COORD_ARRAY)
        self.drawFlash()
        glEnableClientState(GL_TEXTURE_COORD_ARRAY)

        glPopMatrix()

        pygame.display.flip()


    def drawThings(self):

        # Compute directional offset of billboard corners.
        off = [math.cos((self.cam[3]-90.0)/180.0*math.pi)*0.5,
                math.sin((self.cam[3]-90.0)/180.0*math.pi)*0.5]

        camvec = [math.cos(self.cam[3]/180.0*math.pi),
                math.sin(self.cam[3]/180.0*math.pi)]

        # Sort sprite array so the transparency will work correctly.
        def camdist(spr):
            rel = [spr.pos[i] - self.cam[i] for i in range(2)]
            return dot(camvec, rel)
        self.sprites = sorted(self.sprites, key=camdist, reverse=True)

        # Construct vertex arrays
        vertices = [] 
        texcoords = []
        for s in self.sprites:
            s.draw(off, vertices, texcoords)
        n = len(vertices)/3
        vertices = glarray(vertices)
        texcoords = glarray(texcoords)
        glDepthMask(GL_FALSE)
        glVertexPointer(3,GL_FLOAT,0,vertices)
        glTexCoordPointer(2,GL_FLOAT,0,texcoords)
        glDrawArrays(GL_QUADS,0,n)
        glDepthMask(GL_TRUE)

        self.sprites = []



    def drawFlash(self):
        if (self.flash[3] > 0.0):
            glColor4fv(self.flash)
            glCallList(self.flashdl)
            self.flash[3] -= 1.0/64


    def playSounds(self):
        self.audio.step(self.cam, self.sounds)
        self.sounds = []



    def loadMap(self, filename):
        # Clear all things.
        self.things = []
        self.things_solid = set()
        self.things_destr = set()
        self.things_reg = set()

        self.map = Map(self, filename)
        self.map.load()

        # Set the PC start position and direction
        if self.map.filename == LEVEL_HUB:
            if self.hubpc['pos'] == None: 
                # Initialize to the start position.
                pos = self.map.start[:2]
                dir = (self.map.start[2],0.0)
            else:
                # Go to level entrance of the last completed level.
                pos = self.hubpc['pos']
                dir = self.hubpc['dir']
                # Only write a save file if a hubpos already exists.
                # This will ensure that the beginning of the game won't save.
                self.writeSave()
        else:
            pos = self.map.start[:2]
            dir = (self.map.start[2],0.0)

        # Set the ammo status. 
        # On exiting a map, this is set by ThingEntrance/Exit.
        # On dying this will be the last set value.
        self.ammo = list(self.prevammo)
        # Reset the keys. Only keys of type 0 stay during a map switch.
        self.mapkeys = [list(self.prevhubkeys)] + [0]*4
        self.audio.playMusic(self.map.music)
        self.pc = ThingPc(self, pos, dir)
        self.boss = None


    def writeSave(self):
        # Write save file
        f = open(FN_SAVE, 'w')
        for i in self.mapkeys[0]:
            f.write('hubkey ' + str(i[0]) + ' ' + str(i[1]) + '\n')
        for i in self.finmaps:
            f.write('finmap ' + i + '\n')
        for i in range(len(self.ammo)):
            f.write('ammo ' + str(i) + ' ' + str(self.ammo[i]) + '\n')
        f.write('hard ' + str(self.hardmode) + '\n')
        f.close()



