
import random
import math

from define import *
from Sprite import Sprite


# General Thing class, used for in-game objects.
class Thing:


    # Minimum approach distance for two things to be considered touching.
    # Used for collision checking with solids.
    APPROACH = 1.0/64


    # __init__ function which should be inherited by all children.
    def __init__(self, main, pos=[0.0, 0.0, 0.0], dir=[0.0, 0.0]):
        self.main = main
        self.main.things.append(self) # Put self in main.things list.
        self.pos = [float(i) for i in pos]
        if len(self.pos) == 2: 
            # When things are given only two position components, it is assumed
            # that they have been created from the map and are moved by half a
            # tile.
            self.pos = [i+0.5 for i in self.pos]
            self.pos.append(0.5)
        self.dir = list(dir)
        if len(self.dir) == 1:
            self.dir.append(0.0)
        # Note that the q-direction (up and down) isn't really used in this 
        # game, so this will usually be 0.
        self.box = [0.25, -0.5, 0.25]
        self.sp = [0.0, 0.0, 0.0]
        self.nrg = 256 # Thing is removed from list when nrg <= 0


    def step(self):
        pass


    # Bound the speed and direction of the thing.
    def limitDirSpeed(self):
        tmpdirp = math.atan2(self.sp[1], self.sp[0])
        tmpdirq = math.atan2(self.sp[2], 
                math.sqrt(self.sp[0]**2 + self.sp[1]**2))
        tmpsp = math.sqrt(self.sp[0]**2 + self.sp[1]**2 + self.sp[2]**2)
        tmpsp -= self.FRIC
        if (tmpsp > self.SP): 
            tmpsp = self.SP
        elif (tmpsp < 0):
            tmpsp = 0
        tmpsp2 = tmpsp * math.cos(tmpdirq)
        self.sp[2] = tmpsp * math.sin(tmpdirq)
        self.sp[0] = tmpsp2 * math.cos(tmpdirp)
        self.sp[1] = tmpsp2 * math.sin(tmpdirp)
        self.dir[0] = self.dir[0] % (2*math.pi)
        self.dir[1] = bound(self.dir[1], -math.pi/2, math.pi/2)


    # Bound only the speed of the thing.
    def limitSpeed(self, maxsp):
        d = math.atan2(self.sp[1], self.sp[0])
        s = math.sqrt(self.sp[0]**2 + self.sp[1]**2)
        s = min(s, maxsp)
        self.sp[0] = s*math.cos(d)
        self.sp[1] = s*math.sin(d)


    # Move the thing according to speed. Ignores solids.
    def move(self):
        self.pos[2] += self.sp[2]
        self.pos[1] += self.sp[1]
        self.pos[0] += self.sp[0]


    # Check whether the thing can be in this position without colliding
    # with a solid.
    # <solidlev> is the minimal solid level we check (see Tile.py).
    # <excluding> is the set of solid things we will ignore in solid checking,
    # for example, if an enemy shoots a projectile, this projectile should not
    # collide with the enemy.
    def placeFree(self, pos, solidlev=1, excluding=set()):
        x0 = int(pos[0]-self.box[0])
        x1 = int(pos[0]+self.box[0])
        y0 = int(pos[1]-self.box[0])
        y1 = int(pos[1]+self.box[0])
        map = self.main.map
        # Check collisions with map.
        for y in range(y0, y1+1):
            for x in range(x0, x1+1):
                if map.getSolid([x,y]) >= solidlev:
                    return False
        # Check collisions with solid things.
        if 'all' not in excluding:
            for t in self.main.things_solid - excluding:
                if self.collide(t,pos):
                    return False
        return True


    # Move the thing according to its speed, taking solids into account.
    # Returns whether a solid was hit.
    # Excluding contains a set of which solid things not to check in collision,
    # for example a projectile should not check collision with its shooter.
    # If the string 'all' is in excluding, no solid things will be checked.
    def moveSolid(self, solidlev=1, excluding=None):
        if excluding == None: 
            excluding = set()
        excluding.add(self)
        hit = False
        for d in range(3):
            pos = list(self.pos)
            pos[d] += self.sp[d]
            if (self.placeFree(pos, solidlev, excluding)):
                self.pos[d] += self.sp[d]
            else:
                hit = True
                while abs(self.sp[d]) > self.APPROACH:
                    self.sp[d] /= 2
                    pos[d] = self.pos[d] + self.sp[d]
                    if (self.placeFree(pos, solidlev, excluding)):
                        self.pos[d] += self.sp[d]
                self.sp[d] = 0.0
        return hit


    # Like moveSolid, except bounces from the solid
    def bounceSolid(self, solidlev=1, excluding=None):
        if excluding == None: 
            excluding = set()
        excluding.add(self)
        hit = False
        pos = list(self.pos)
        for d in range(3):
            pos[d] += self.sp[d]
            if (self.placeFree(pos, solidlev, excluding)):
                self.pos[d] += self.sp[d]
            else: 
                hit = True
                self.sp[d] *= -1
                return hit
        return hit


    # Returns whether two things collide at their current positions.
    # If pos != None, it will check for self moved to position pos.
    def collide(self, other, pos=None): 
        if pos == None:
            return (not 
                    (self.pos[0] - self.box[0] > other.pos[0] + other.box[0] or
                    self.pos[0] + self.box[0] < other.pos[0] - other.box[0] or
                    self.pos[1] - self.box[0] > other.pos[1] + other.box[0] or
                    self.pos[1] + self.box[0] < other.pos[1] - other.box[0]))
        else:
            return (not 
                    (pos[0] - self.box[0] > other.pos[0] + other.box[0] or
                    pos[0] + self.box[0] < other.pos[0] - other.box[0] or
                    pos[1] - self.box[0] > other.pos[1] + other.box[0] or
                    pos[1] + self.box[0] < other.pos[1] - other.box[0]))


    # Returns list of destructable things this Thing collides with.
    def collideDestr(self):
        l = []
        for t in self.main.things_destr:
            if self.collide(t): 
                l.append(t)
        return l


    # Returns whether the thing is at most dist away from a target position.
    def isNearTarget(self, dist):
        return ((self.pos[0]-self.target[0])**2 +
                (self.pos[1]-self.target[1])**2 <= dist**2)


    # Set acceleration and direction towards a target position.
    def chaseTarget(self, acc):
        d = self.getTargetDir()
        self.sp[0] += acc*math.cos(d)
        self.sp[1] += acc*math.sin(d)


    # Set current target to PC
    def setPcTarget(self):
        self.target = self.main.pc.pos


    # Return the target direction
    def getTargetDir(self):
        return math.atan2(self.target[1] - self.pos[1], 
                self.target[0] - self.pos[0])

    # Return the PC direction
    def getPcDir(self):
        return math.atan2(self.main.pc.pos[1] - self.pos[1], 
                self.main.pc.pos[0] - self.pos[0])


    # Set target to random position from current
    def setRandomTarget(self, dist=8):
        self.target = [self.pos[i] + random.uniform(-dist, dist) 
                for i in range(2)]


    # Returns whether the thing can see the PC
    # Not the lightest function, so I should avoid calling it too often.
    def canSeePc(self):
        return self.canSee(self.main.pc)


    # Returns whether the thing can see the other thing.
    # Draws a line from self to other and checks whether it collides with solid 
    # tiles anywhere.
    # Note that this uses points for both things, it is still possible to see 
    # edge of a things sprite without this returning True.
    def canSee(self, other):
        x0 = self.pos[0]
        y0 = self.pos[1]
        x1 = other.pos[0]
        y1 = other.pos[1]
        map = self.main.map
        if x0 > x1: x0, y0, x1, y1 = x1, y1, x0, y0
        if int(x0) == int(x1):
            ya = y0
            yb = y1
            if yb < ya: ya, yb = yb, ya
            for y in range(int(ya),int(yb)+1):
                pos = [int(x0), y]
                if map.getSolid([int(x0), y]) >= 3:
                    return False
        else:
            b = (y1-y0)/(x1-x0)
            ya = y0
            yb = y0+b*(int(x0)-x0+1)
            if yb < ya: ya, yb = yb, ya
            for y in range(int(ya),int(yb)+1):
                pos = [int(x0), y]
                if map.getSolid(pos) >= 3:
                    return False
            for x in range(int(x0)+1,int(x1)):
                ya = y0+b*(x-x0)
                yb = y0+b*(x-x0+1)
                if yb < ya: ya, yb = yb, ya
                for y in range(int(ya),int(yb)+1):
                    pos = [x, y]
                    if map.getSolid(pos) >= 3:
                        return False
            ya = y0+b*(int(x1)-x0)
            yb = y1
            if yb < ya: ya, yb = yb, ya
            for y in range(int(ya),int(yb)+1):
                pos = [int(x1), y]
                if map.getSolid(pos) >= 3:
                    return False
        return True


    # Returns whether the thing is in the camera view. Ignores solids.
    def inCamView(self):
        cam = self.main.cam
        cd = cam[3]/180.0*math.pi
        sd = math.atan2(self.pos[1]-cam[1], self.pos[0]-cam[0])
        vs = (self.pos[0]-cam[0], self.pos[1]-cam[1])
        hfov = math.pi/4
        v0 = (math.cos(cd-hfov), math.sin(cd-hfov))
        v1 = (math.cos(cd+hfov), math.sin(cd+hfov))
        return (dot(vs, v0) >= -0.5 and dot(vs, v1) > -0.5)


