
import math
import random

from Thing import Thing
from Sprite import Sprite
from ThingBossProj import ThingBossProj
from ThingDragonProj import ThingDragonProj
from ThingGhost import ThingGhost
from ThingDeadBoss import ThingDeadBoss
from ThingNova import ThingNova
from ThingBossSpray import ThingBossSpray
from Sfx import Sfx


# Final boss. Flies around in circles and throws attacks at the PC.
class ThingBoss(Thing):


    # Flying circle radius
    RADIUS = 8.0
    # Projectile patterns, 1 means a projectile will be placed at this time 
    # step. One of these will be selected at random. Maximal length of these 
    # should be 15 or so.
    PROJ_PATTERNS = [
            [1,0,1,0,1],
            [1,0,0,0,1],
            [1,0,0,1,1,0,0,1],
            [1,0,1,0,0,1,0,1],
            [1,1,1,0,0,1,1,1],
            [1,1,0,1,1,0,1,1],
            [1,0,0,1,0,0,1],
            [1,1,1,1,1,1,1,1,1]
    ]
    # Movement acceleration.
    ACC = math.pi/4096
    # Speed approach distance.
    SP_MINDIST = math.pi/4096
    # Fire attack spread
    FIRE_SPREAD = math.pi/12
    # Where to shoot fireballs.
    FIRE_RANGE = range(-2,3)
    # What patterns to fire in, range given above.
    FIRE_PATTERNS = [
        [0,1,0,1,0],
        [1,0,1,0,1],
        [1,1,0,1,1],
        [0,1,1,1,0],
        [1,1,1,1,1]
    ]
    # Maximal amount of ghosts that can be spawned.
    MAXGHOSTS = 16
    # Probability of ghost spawning each step (1/p)
    GHOST_P = 32
    # Spray attack interval
    SPRAY_P = 6
    # Fire attack interval
    FIRE_INTERVAL = 96


    # args: [pos_x, pos_y, flag_id]
    def __init__(self, main, args):
        Thing.__init__(self, main, args[:2])
        self.center = list(self.pos)
        self.flag = int(args[2])
        self.box[0] = 0.375
        self.nrg = 200
        self.state = 0
        self.timer = 0
        # This enemy behaves a bit different, since its position is calculated 
        # from its center position, circle radius and angle (theta)
        self.theta = 0.0
        self.tsp = math.pi/32.0 # Theta speed
        self.goalsp = math.pi/256.0 # Speed we want to reach
        # Firing pattern
        self.pattern = []
        # Add flag 99 to the map so the ghosts will spawn instantaneously.
        self.main.map.flags.add(99)
        self.ghosts = []
        if self.main.hardmode:
            self.GHOST_P = 16
            self.SPRAY_P = 4
            self.FIRE_SPREAD = math.pi/16
            self.FIRE_INTERVAL = 48

    
    def step(self):
        if self.nrg <= 0: return
        self.STATEFUNCTIONS[self.state](self)
        if self.main.hardmode:
            self.timer -= 2
        else:
            self.timer -= 1


    def damage(self, nrg):
        if self.state != 5 and self.nrg > 0:
            self.nrg -= nrg
            if self.nrg <= 0:
                ThingDeadBoss(self.main, self.pos)
                Sfx(self.main, 'boss_hurt.ogg', self, 2500.0)
                Sfx(self.main, 'nova_explode.ogg', None)
                self.main.boss = None
            else:
                if random.randint(0,3) == 0:
                    Sfx(self.main, 'boss_hurt.ogg', self, 2500.0)


    # Update the position according to the Theta speed value.
    def updatePos(self):
        self.theta -= self.tsp
        self.theta %= 2*math.pi
        self.pos[0] = self.center[0] + self.RADIUS*math.cos(self.theta)
        self.pos[1] = self.center[1] + self.RADIUS*math.sin(self.theta)
        self.setPcTarget()


    # Sprite drawn depends on direction.
    def drawSprite(self):
        if self.tsp < 0: 
            spr = (12,1) 
        else: 
            spr = (13,1)
        Sprite(self.main, self.pos, spr)


    # Pick a random next attack state.
    def setStateNext(self):
        r = random.randint(0,9)
        if r in [0,1,2]:
            self.setStateAcc(random.uniform(math.pi/1024, math.pi/128)*
                    random.choice([-1,1]), self.setStateSphere)
        elif r in [3,4]:
            self.setStateAcc(random.uniform(math.pi/1024, math.pi/256)*
                    random.choice([-1,1]), self.setStateFire)
        elif r in [5,6]:
            self.setStateAcc(random.uniform(math.pi/1024, math.pi/256)*
                    random.choice([-1,1]), self.setStateNova)
        elif r in [7,8]: 
            self.setStateAcc(random.uniform(math.pi/1024, math.pi/256)*
                    random.choice([-1,1]), self.setStateSpray)
        else:
            Sfx(self.main, 'boss_activate.ogg', self, 2500.0)
            self.setStateAcc(math.pi/16*random.choice([-1,1]), 
                    self.setStateHide)



    def setStateAcc(self, sp, nextfun):
        self.goalsp = sp
        self.state = 1
        self.nextfun = nextfun
        self.timer = 64


    def setStateSphere(self):
        self.pattern = random.choice(self.PROJ_PATTERNS)
        self.timer = 256
        self.state = 3


    def setStateFire(self):
        self.pattern = random.choice(self.FIRE_PATTERNS)
        self.timer = 256
        self.state = 4


    def setStateNova(self):
        self.timer = 256
        self.state = 6


    def setStateHide(self):
        self.timer = 256
        self.state = 5


    def setStateSpray(self):
        self.timer = 256
        self.state = 7


    def stepStateFlag(self):
        # Waiting for flag
        if self.flag in self.main.map.flags:
            Sfx(self.main, 'boss_activate.ogg', self, 8.0)
            self.setStateNext()
            self.main.things_solid.add(self)
            self.main.things_destr.add(self)
            self.main.boss = self


    def stepStateAcc(self):
        neargoal = abs(self.tsp-self.goalsp) < self.SP_MINDIST
        if not neargoal:
            if self.goalsp > self.tsp:
                self.tsp += self.ACC
            else:
                self.tsp -= self.ACC
        if self.timer <= 0 and neargoal:
            self.nextfun()
        self.updatePos()
        self.drawSprite()


    # Sphere state, will drop a series of spheres in the current pattern.
    def stepStateSphere(self):
        self.updatePos()
        t = 256-self.timer
        if t%16 == 0 and len(self.pattern) > t/16 and self.pattern[t/16] == 1:
            ThingBossProj(self.main, self.pos, self.target, self)
        if self.timer <= 0:
            self.setStateNext()
        self.drawSprite()


    # Will blast a few (Dragon) fire shots in the current pattern.
    def stepStateFire(self):
        self.updatePos()
        if self.timer%self.FIRE_INTERVAL == 32:
            self.setPcTarget()
            d = self.getTargetDir()
            d = [d+i*self.FIRE_SPREAD for i in self.FIRE_RANGE if 
                    self.pattern[i-max(self.FIRE_RANGE)-1]]
            Sfx(self.main, 'dragon_fire.ogg', self)
            for i in d:
                ThingDragonProj(self.main, self.pos, [i], self)
        if self.timer <= 0:
            self.setStateNext()
        self.drawSprite()


    # Will create a Nova projectile and throw it at the PC.
    def stepStateNova(self):
        self.updatePos()
        if self.timer == 192:
            ThingNova(self.main, self.pos, self.target, self)
        if self.timer <= 0:
            self.setStateNext()
        self.drawSprite()


    # Will be invisible while spawning Ghosts.
    def stepStateHide(self):
        self.updatePos()
        if (self.timer > 128 and random.randint(0,self.GHOST_P-1) == 0 and 
                len(self.ghosts) < self.MAXGHOSTS):
            self.ghosts.append(
                    ThingGhost(self.main, [self.pos[0], self.pos[1], 99]))
        if self.timer <= 0:
            self.setStateNext()


    # Spray attack, will unleash a spray of projectiles. Only used as "revenge"
    # for the TV attack until low on health.
    def stepStateSpray(self):
        self.updatePos()
        if self.timer > 64 and self.main.step_i % self.SPRAY_P == 0:
            d = self.getTargetDir()
            ThingBossSpray(self.main, self.pos, [d+random.uniform(-0.5,0.5)],
                    self)
        if self.timer <= 0:
            self.setStateNext()
        self.drawSprite()


    STATEFUNCTIONS = {
            0: stepStateFlag,
            1: stepStateAcc,
            3: stepStateSphere,
            4: stepStateFire,
            5: stepStateHide,
            6: stepStateNova,
            7: stepStateSpray
    }

