

import math
import random

from Thing import Thing
from ThingAni import ThingAni
from Sprite import Sprite
from ThingHurt import ThingHurt
from ThingDragonProj import ThingDragonProj
from Sfx import Sfx


class ThingDragon(Thing):


    ACC_X = 1.0/128
    SP_X = 1.0/16
    SP_Z = 1.0/16
    PROJ_SPREAD = math.pi/8


    # args: [pos_x, pos_y, flag_id]
    # If the trigger id is negative, it will be ignored. Otherwise the enemy
    # will only pop up if the corresponding trigger has been hit.
    def __init__(self, main, args):
        Thing.__init__(self, main, args[:2])
        self.flag = int(args[2])
        self.box[0] = 0.375
        self.nrg = 15
        if self.main.hardmode:
            self.nrg = 30
        self.state = 0
        if self.flag < 0: self.state = 1
        self.pos[2] = -0.5
        self.timer = 0
        self.ani_offset = random.randint(0,31)


    # Function called when damage is inflicted on the Thing. 
    def damage(self, nrg):
        if self.state < 3: return
        self.nrg -= nrg
        if self.nrg <= 0:
            ThingAni(self.main, self.pos, (0,7), 5, 8)
            Sfx(self.main, 'dragon_death.ogg', self)
        if random.randint(0,3) == 0:
            self.state = 6
            self.setRandomTarget()
            self.timer = 15


    def step(self):
        if self.nrg <= 0: return
        self.STATEFUNCTIONS[self.state](self)
        self.timer -= 1


    def getWalkSprite(self):
        x = (self.main.step_i+self.ani_offset) % 32 / 8
        return (x,6)


    def setStatePopup(self):
        if not self.placeFree(self.pos, 2, set([self])):
            return
        self.state = 2
        self.sp[2] = self.SP_Z
        self.main.things_solid.add(self)
        self.main.things_destr.add(self)
        Sfx(self.main, 'dragon_activate.ogg', self, 2.0)


    def setStateChase(self):
        self.state = 3
        self.pos[2] = 0.5
        self.sp[2] = 0.0
        self.timer = 300


    def setStateRetreat(self, time=150):
        self.state = 4
        self.setRandomTarget()
        self.timer = time


    def setStateAttack(self, time=15):
        self.state = 5
        self.timer = time


    def moveToTarget(self):
        self.chaseTarget(self.ACC_X)
        self.limitSpeed(self.SP_X)
        self.posprev = list(self.pos)
        self.moveSolid(2)


    def stepStateFlag(self):
        # Waiting for flag
        if self.flag in self.main.map.flags:
            self.setStatePopup()


    def stepStateWait(self):
        # Waiting for PC to approach.
        self.setPcTarget()
        if self.isNearTarget(2.0):
            self.setStatePopup()


    def stepStatePopup(self):
        # Popping up from the ground
        self.move()
        if self.pos[2] >= 0.5: 
            self.setStateChase()
        Sprite(self.main, self.pos, self.getWalkSprite())


    def stepStateChase(self):
        # Chasing the PC
        self.setPcTarget()
        self.moveToTarget()
        if self.timer <= 0:
            # Timeout, retreat a bit
            self.setStateRetreat(random.randint(60,120))
        if self.posprev == self.pos and self.sp == [0.0, 0.0, 0.0]:
            # Stuck, retreat a bit
            self.setStateRetreat(random.randint(60,120))
        if self.isNearTarget(8.0) and self.canSeePc():
            # Near PC, attack
            self.setStateAttack()
        if (self.main.hardmode and random.randint(0,127) == 0 and 
                self.canSeePc()):
            Sfx(self.main, 'dragon_fire.ogg', self)
            d = self.getPcDir()
            ThingDragonProj(self.main, self.pos, [d], self)
        Sprite(self.main, self.pos, self.getWalkSprite())


    def stepStateRetreat(self):
        # Retreating
        self.moveToTarget()
        if self.posprev == self.pos and self.sp == [0.0, 0.0, 0.0]:
            self.setRandomTarget()
        if self.timer <= 0:
            self.setStateChase()
        if (self.main.hardmode and random.randint(0,127) == 0 and 
                self.canSeePc()):
            Sfx(self.main, 'dragon_fire.ogg', self)
            d = self.getPcDir()
            ThingDragonProj(self.main, self.pos, [d], self)
        Sprite(self.main, self.pos, self.getWalkSprite())


    def stepStateAttack(self):
        # Attacking (firing)
        self.setPcTarget()
        d = self.getTargetDir()
        self.moveToTarget()
        if self.timer <= 0:
            Sfx(self.main, 'dragon_fire.ogg', self)
            ThingDragonProj(self.main, self.pos, [d-self.PROJ_SPREAD], self)
            ThingDragonProj(self.main, self.pos, [d+self.PROJ_SPREAD], self)
            ThingDragonProj(self.main, self.pos, [d], self)
            self.setStateRetreat(random.randint(120,240))
        Sprite(self.main, self.pos, self.getWalkSprite())


    def stepStateHurt(self):
        # Hurt
        self.moveToTarget()
        if self.posprev == self.pos and self.sp == [0.0, 0.0, 0.0]:
            self.setRandomTarget()
        if self.timer <= 0:
            self.setStateRetreat(random.randint(60,120))
        Sprite(self.main, self.pos, (0,7))


    STATEFUNCTIONS = {
            0: stepStateFlag,
            1: stepStateWait,
            2: stepStatePopup,
            3: stepStateChase,
            4: stepStateRetreat,
            5: stepStateAttack,
            6: stepStateHurt}



